@@ -58,7 +58,6 @@ class WP_REST_Server {
 	 * Namespaces registered to the server.
 	 *
 	 * @since 4.4.0
-	 * @access protected
 	 * @var array
 	 */
 	protected $namespaces = array();
@@ -67,7 +66,6 @@ class WP_REST_Server {
 	 * Endpoints registered to the server.
 	 *
 	 * @since 4.4.0
-	 * @access protected
 	 * @var array
 	 */
 	protected $endpoints = array();
@@ -76,7 +74,6 @@ class WP_REST_Server {
 	 * Options defined for the routes.
 	 *
 	 * @since 4.4.0
-	 * @access protected
 	 * @var array
 	 */
 	protected $route_options = array();
@@ -85,7 +82,6 @@ class WP_REST_Server {
 	 * Instantiates the REST server.
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 */
 	public function __construct() {
 		$this->endpoints = array(
@@ -107,7 +103,6 @@ class WP_REST_Server {
 	 * Checks the authentication headers if supplied.
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 *
 	 * @return WP_Error|null WP_Error indicates unsuccessful login, null indicates successful
 	 *                       or no authentication provided
@@ -148,7 +143,6 @@ class WP_REST_Server {
 	 * list in JSON rather than an object/map.
 	 *
 	 * @since 4.4.0
-	 * @access protected
 	 *
 	 * @param WP_Error $error WP_Error instance.
 	 * @return WP_REST_Response List of associative arrays with code and message keys.
@@ -191,7 +185,6 @@ class WP_REST_Server {
 	 * a 'status' key, with the value being the HTTP status to send.
 	 *
 	 * @since 4.4.0
-	 * @access protected
 	 *
 	 * @param string $code    WP_Error-style code.
 	 * @param string $message Human-readable message.
@@ -215,7 +208,6 @@ class WP_REST_Server {
 	 * callback then outputs a JSON representation of the returned value.
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 *
 	 * @see WP_REST_Server::dispatch()
 	 *
@@ -250,7 +242,9 @@ class WP_REST_Server {
 		 *
 		 * @param bool $rest_enabled Whether the REST API is enabled. Default true.
 		 */
-		apply_filters_deprecated( 'rest_enabled', array( true ), '4.7.0', 'rest_authentication_errors', __( 'The REST API can no longer be completely disabled, the rest_authentication_errors can be used to restrict access to the API, instead.' ) );
+		apply_filters_deprecated( 'rest_enabled', array( true ), '4.7.0', 'rest_authentication_errors',
+			__( 'The REST API can no longer be completely disabled, the rest_authentication_errors filter can be used to restrict access to the API, instead.' )
+		);
 
 		/**
 		 * Filters whether jsonp is enabled.
@@ -375,7 +369,11 @@ class WP_REST_Server {
 		// OR if the HTTP_X_HTTP_METHOD_OVERRIDE is used but resulted a 4xx response code.
 		if ( $send_no_cache_headers || ( true === $method_overridden && strpos( $code, '4' ) === 0 ) ) {
 			foreach ( wp_get_nocache_headers() as $header => $header_value ) {
-				$this->send_header( $header, $header_value );
+				if ( empty( $header_value ) ) {
+					$this->remove_header( $header );
+				} else {
+					$this->send_header( $header, $header_value );
+				}
 			}
 		}
 
@@ -387,6 +385,20 @@ class WP_REST_Server {
 			// Embed links inside the request.
 			$result = $this->response_to_data( $result, isset( $_GET['_embed'] ) );
 
+			/**
+			 * Filters the API response.
+			 *
+			 * Allows modification of the response data after inserting
+			 * embedded data (if any) and before echoing the response data.
+			 *
+			 * @since 4.8.1
+			 *
+			 * @param array            $result  Response data to send to the client.
+			 * @param WP_REST_Server   $this    Server instance.
+			 * @param WP_REST_Request  $request Request used to generate the response.
+			 */
+			$result = apply_filters( 'rest_pre_echo_response', $result, $this, $request );
+
 			$result = wp_json_encode( $result );
 
 			$json_error_message = $this->get_json_last_error();
@@ -411,7 +423,6 @@ class WP_REST_Server {
 	 * Converts a response to data to send.
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 *
 	 * @param WP_REST_Response $response Response object.
 	 * @param bool             $embed    Whether links should be embedded.
@@ -449,7 +460,6 @@ class WP_REST_Server {
 	 * direct output.
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 * @static
 	 *
 	 * @param WP_REST_Response $response Response to extract links from.
@@ -483,7 +493,6 @@ class WP_REST_Server {
 	 * direct output.
 	 *
 	 * @since 4.5.0
-	 * @access public
 	 * @static
 	 *
 	 * @param WP_REST_Response $response Response to extract links from.
@@ -533,7 +542,6 @@ class WP_REST_Server {
 	 * Embeds the links from the data into the request.
 	 *
 	 * @since 4.4.0
-	 * @access protected
 	 *
 	 * @param array $data Data from the request.
 	 * @return array {
@@ -609,7 +617,6 @@ class WP_REST_Server {
 	 * data instead.
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 *
 	 * @param WP_REST_Response $response Response object.
 	 * @param bool             $embed    Whether links should be embedded.
@@ -640,7 +647,6 @@ class WP_REST_Server {
 	 * Registers a route to the server.
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 *
 	 * @param string $namespace  Namespace.
 	 * @param string $route      The REST route.
@@ -695,7 +701,6 @@ class WP_REST_Server {
 	 * used as the delimiter with preg_match()
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 *
 	 * @return array `'/path/regex' => array( $callback, $bitmask )` or
 	 *               `'/path/regex' => array( array( $callback, $bitmask ), ...)`.
@@ -770,7 +775,6 @@ class WP_REST_Server {
 	 * Retrieves namespaces registered on the server.
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 *
 	 * @return array List of registered namespaces.
 	 */
@@ -782,7 +786,6 @@ class WP_REST_Server {
 	 * Retrieves specified options for a route.
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 *
 	 * @param string $route Route pattern to fetch options for.
 	 * @return array|null Data as an associative array if found, or null if not found.
@@ -799,7 +802,6 @@ class WP_REST_Server {
 	 * Matches the request to a callback and call it.
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 *
 	 * @param WP_REST_Request $request Request to attempt dispatching.
 	 * @return WP_REST_Response Response returned by the callback.
@@ -828,12 +830,19 @@ class WP_REST_Server {
 		$path   = $request->get_route();
 
 		foreach ( $this->get_routes() as $route => $handlers ) {
-			$match = preg_match( '@^' . $route . '$@i', $path, $args );
+			$match = preg_match( '@^' . $route . '$@i', $path, $matches );
 
 			if ( ! $match ) {
 				continue;
 			}
 
+			$args = array();
+			foreach ( $matches as $param => $value ) {
+				if ( ! is_int( $param ) ) {
+					$args[ $param ] = $value;
+				}
+			}
+
 			foreach ( $handlers as $handler ) {
 				$callback  = $handler['callback'];
 				$response = null;
@@ -905,7 +914,7 @@ class WP_REST_Server {
 						if ( is_wp_error( $permission ) ) {
 							$response = $permission;
 						} elseif ( false === $permission || null === $permission ) {
-							$response = new WP_Error( 'rest_forbidden', __( 'Sorry, you are not allowed to do that.' ), array( 'status' => 403 ) );
+							$response = new WP_Error( 'rest_forbidden', __( 'Sorry, you are not allowed to do that.' ), array( 'status' => rest_authorization_required_code() ) );
 						}
 					}
 				}
@@ -979,7 +988,6 @@ class WP_REST_Server {
 	 * "Encoding error: Maximum stack depth exceeded".
 	 *
 	 * @since 4.4.0
-	 * @access protected
 	 *
 	 * @return bool|string Boolean false or string error message.
 	 */
@@ -1004,7 +1012,6 @@ class WP_REST_Server {
 	 * This endpoint describes the capabilities of the site.
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 *
 	 * @param array $request {
 	 *     Request.
@@ -1049,7 +1056,6 @@ class WP_REST_Server {
 	 * Retrieves the index for a namespace.
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 *
 	 * @param WP_REST_Request $request REST request instance.
 	 * @return WP_REST_Response|WP_Error WP_REST_Response instance if the index was found,
@@ -1092,7 +1098,6 @@ class WP_REST_Server {
 	 * Retrieves the publicly-visible data for routes.
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 *
 	 * @param array  $routes  Routes to get data for.
 	 * @param string $context Optional. Context for data. Accepts 'view' or 'help'. Default 'view'.
@@ -1137,7 +1142,6 @@ class WP_REST_Server {
 	 * Retrieves publicly-visible data for the route.
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 *
 	 * @param string $route     Route to get data for.
 	 * @param array  $callbacks Callbacks to convert to data.
@@ -1223,7 +1227,6 @@ class WP_REST_Server {
 	 * Sends an HTTP status code.
 	 *
 	 * @since 4.4.0
-	 * @access protected
 	 *
 	 * @param int $code HTTP status.
 	 */
@@ -1235,7 +1238,6 @@ class WP_REST_Server {
 	 * Sends an HTTP header.
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 *
 	 * @param string $key Header key.
 	 * @param string $value Header value.
@@ -1256,7 +1258,6 @@ class WP_REST_Server {
 	 * Sends multiple HTTP headers.
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 *
 	 * @param array $headers Map of header name to header value.
 	 */
@@ -1266,11 +1267,33 @@ class WP_REST_Server {
 		}
 	}
 
+	/**
+	 * Removes an HTTP header from the current response.
+	 *
+	 * @since 4.8.0
+	 *
+	 * @param string $key Header key.
+	 */
+	public function remove_header( $key ) {
+		if ( function_exists( 'header_remove' ) ) {
+			// In PHP 5.3+ there is a way to remove an already set header.
+			header_remove( $key );
+		} else {
+			// In PHP 5.2, send an empty header, but only as a last resort to
+			// override a header already sent.
+			foreach ( headers_list() as $header ) {
+				if ( 0 === stripos( $header, "$key:" ) ) {
+					$this->send_header( $key, '' );
+					break;
+				}
+			}
+		}
+	}
+
 	/**
 	 * Retrieves the raw request entity (body).
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 *
 	 * @global string $HTTP_RAW_POST_DATA Raw post data.
 	 *
@@ -1294,7 +1317,6 @@ class WP_REST_Server {
 	 * Extracts headers from a PHP-style $_SERVER array.
 	 *
 	 * @since 4.4.0
-	 * @access public
 	 *
 	 * @param array $server Associative array similar to `$_SERVER`.
 	 * @return array Headers extracted from the input.
